use anyhow::{bail, Result};

use crate::data::{Context, ContextEntry, FreeVar, Monotype, Term, Type};
use crate::gensym::gensym_existential;

// Figure 8. Applying a context, as a substitution, to a type

#[cfg_attr(feature = "trace_execution", trace)]
pub fn app_ctx(ctx: &Context, ty: &Type) -> Result<Type> {
  match ty {
    Type::Unit => Ok(Type::Unit),
    Type::Var(s) => Ok(Type::Var(s.clone())),

    Type::Existential(a) => match ctx.lookup_existential(a) {
      Some((_, Some(m))) => Ok(m.into_poly()),
      Some((_, None)) => Ok(Type::Existential(a.clone())),
      None => bail!("existential variable {a} doesn't exist in context"),
    },

    Type::Polytype(a, t) => {
      Ok(Type::Polytype(a.clone(), Box::new(app_ctx(ctx, t)?)))
    }

    Type::Arrow(a, b) => Ok(Type::Arrow(
      Box::new(app_ctx(ctx, a)?),
      Box::new(app_ctx(ctx, b)?),
    )),
  }
}

// Figure 9. Algorithmic subtyping

/// Under input context Γ , type A is a subtype of B, with output context ∆
#[cfg_attr(feature = "trace_execution", trace)]
pub fn subtype(ctx: &Context, left: &Type, right: &Type) -> Result<Context> {
  match (left, right) {
    // <:Unit rule
    (Type::Unit, Type::Unit) => Ok(ctx.clone()),

    // <:Var rule
    (Type::Var(x), Type::Var(y)) if x == y && ctx.lookup_type(x).is_some() => Ok(ctx.clone()),

    // <:Exvar rule
    (Type::Existential(x), Type::Existential(y))
      if x == y && ctx.lookup_existential(x).is_some() =>
    {
      Ok(ctx.clone())
    }

    // <:InstantiateL
    (Type::Existential(a), ty_a)
      if !ty_a
        .free_vars()
        .contains(&FreeVar::Existential(a.to_string()))
        && ctx.lookup_existential(a).is_some() =>
    {
      let ctx_delta = instantiate_left(ctx, a, ty_a)?;
      Ok(ctx_delta)
    }

    // <:InstantiateR
    (ty_a, Type::Existential(a))
      if !ty_a
        .free_vars()
        .contains(&FreeVar::Existential(a.to_string()))
        && ctx.lookup_existential(a).is_some() =>
    {
      let ctx_delta = instantiate_right(ctx, ty_a, a)?;
      Ok(ctx_delta)
    }

    (Type::Existential(_), Type::Unit) => todo!(),
    (Type::Existential(_), Type::Var(_)) => todo!(),
    (Type::Existential(_), Type::Polytype(_, _)) => todo!(),
    (Type::Existential(_), Type::Arrow(_, _)) => todo!(),
    (Type::Polytype(_, _), Type::Unit) => todo!(),
    (Type::Polytype(_, _), Type::Var(_)) => todo!(),
    (Type::Polytype(_, _), Type::Existential(_)) => todo!(),
    (Type::Polytype(_, _), Type::Polytype(_, _)) => todo!(),
    (Type::Polytype(_, _), Type::Arrow(_, _)) => todo!(),
    (Type::Arrow(_, _), Type::Unit) => todo!(),
    (Type::Arrow(_, _), Type::Var(_)) => todo!(),
    (Type::Arrow(_, _), Type::Existential(_)) => todo!(),
    (Type::Arrow(_, _), Type::Polytype(_, _)) => todo!(),
    (Type::Arrow(_, _), Type::Arrow(_, _)) => todo!(),

    _ => bail!("subtyping relation failed between {left:?} and {right:?} (ctx = {ctx:?})"),
  }
}

// Figure 10. Instantiation

#[cfg_attr(feature = "trace_execution", trace)]
pub fn instantiate_left(
  ctx: &Context,
  a: &str,
  ty_a: &Type,
) -> Result<Context> {
  match ty_a {
    // InstLReach rule
    Type::Existential(b)
      if ctx.has_existential(a) && ctx.has_existential(b) =>
    {
      let mut new_ctx = ctx.clone();
      {
        let b_entry = new_ctx
          .0
          .iter_mut()
          .find(
            |entry| matches!(entry, ContextEntry::ExistentialVar(x) if x == b),
          )
          .expect("should exist");

        *b_entry = ContextEntry::ExistentialSolved(
          b.to_owned(),
          Monotype::Existential(a.to_owned()),
        );
      }

      Ok(new_ctx)
    }

    Type::Existential(_b) => todo!(),

    Type::Unit => todo!(),
    Type::Var(_) => todo!(),
    Type::Polytype(_, _) => todo!(),
    Type::Arrow(_, _) => todo!(),
  }
}

#[cfg_attr(feature = "trace_execution", trace)]
pub fn instantiate_right(
  ctx: &Context,
  ty_a: &Type,
  a: &str,
) -> Result<Context> {
  match ty_a {
    // InstRArr
    Type::Arrow(ty_a1, ty_a2) if ctx.has_existential(a) => {
      // TODO: In case it's already solved?
      println!("original ctx: {ctx:?}");

      let (ex_a1_s, ex_a2_s, ctx_gamma_aug) =
        ctx.split_existential_function(a)?;

      let ctx_theta = instantiate_left(&ctx_gamma_aug, &ex_a1_s, &ty_a1)?;

      let ty_a2_aug = app_ctx(&ctx_theta, &ty_a2)?;
      let ctx_delta = instantiate_right(&ctx_theta, &ty_a2_aug, &ex_a2_s)?;

      Ok(ctx_delta)
    }

    // InstRAllL
    Type::Polytype(beta, ty_b) if ctx.has_existential(a) => {
      let ex_b = gensym_existential();
      let aug_ctx = ctx.add(vec![
        ContextEntry::Marker(ex_b.to_owned()),
        ContextEntry::ExistentialVar(ex_b.to_owned()),
      ]);

      let aug_b = ty_b.subst(beta, &Type::Existential(ex_b.to_owned()));
      let out_ctx = instantiate_right(&aug_ctx, &aug_b, a)?;

      let (before, _after) = out_ctx
        .split_by(
          |entry| matches!(entry, ContextEntry::Marker(m) if *m == ex_b),
        )
        .unwrap();

      Ok(before)
    }

    // InstRSolve
    ty_a if ty_a.is_mono() && ctx.has_existential(a) => {
      let (before, after) = ctx
        .split_by(|entry| match entry {
          ContextEntry::ExistentialVar(n) if n == a => true,
          _ => false,
        })
        .unwrap();

      let ty_a_mono = ty_a.into_mono().unwrap();

      Ok(Context::unsplit(
        &before,
        ContextEntry::ExistentialSolved(a.to_owned(), ty_a_mono),
        &after,
      ))
    }

    Type::Unit => todo!(),
    Type::Var(_) => todo!(),
    Type::Existential(_) => todo!(),

    _ => todo!(),
  }
}

// Figure 11. Algorithmic typing

#[cfg_attr(feature = "trace_execution", trace)]
pub fn typecheck(ctx: &Context, term: &Term, ty: &Type) -> Result<Context> {
  match (term, ty) {
    // 1I rule
    (Term::Unit, Type::Unit) => Ok(ctx.clone()),

    // ∀I rule
    (_e, Type::Polytype(_x, _tyA)) => todo!(),

    // →I rule
    (Term::Lam(_x, _e), Type::Arrow(_ty_a, _ty_b)) => {
      let _aug_ctx = ctx.clone();
      todo!()
    }

    // Sub rule
    (term, ty) => {
      let (ty_a, ctx_theta) = synthesize(ctx, term)?;
      let a = app_ctx(&ctx_theta, &ty_a)?;
      let b = app_ctx(&ctx_theta, ty)?;
      let ctx_delta = subtype(&ctx_theta, &a, &b)?;
      Ok(ctx_delta)
    }
  }
}

#[cfg_attr(feature = "trace_execution", trace)]
pub fn synthesize(ctx: &Context, term: &Term) -> Result<(Type, Context)> {
  match term {
    // Var rule
    Term::Var(name) if ctx.has_type(name) => {
      let (_, ty) = ctx.lookup_type(name).unwrap();
      Ok((ty, ctx.clone()))
    }
    Term::Var(name) => bail!("could not find name {name}"),

    // 1I⇒ rule
    Term::Unit => Ok((Type::Unit, ctx.clone())),

    // Anno rule
    Term::Annot(e, ty_a) => {
      // Make sure the type is well formed
      // TODO: Have an actual "well-formed" check

      let ctx_delta = typecheck(ctx, &e, ty_a)?;
      Ok((ty_a.clone(), ctx_delta))
    }

    // →I⇒' rule
    Term::Lam(x, e) => {
      let ex_a_s = gensym_existential();
      let ex_b_s = gensym_existential();

      let ex_gen = format!("{}*", ex_a_s);

      let ex_a = Type::Existential(ex_a_s.clone());
      let ex_b = Type::Existential(ex_b_s.clone());

      let aug_ctx = ctx.add(vec![
        ContextEntry::Marker(ex_a_s.clone()),
        ContextEntry::ExistentialVar(ex_a_s.clone()),
        ContextEntry::ExistentialVar(ex_b_s.clone()),
        ContextEntry::TermAnnot(x.clone(), ex_a.clone()),
      ]);

      let wtf_ctx = typecheck(&aug_ctx, &e, &ex_b)?;
      info!("WTF CTX {wtf_ctx:?}");

      // Split back out at the marker
      let (before_marker, after_marker) = wtf_ctx
        .split_by(|entry| match entry {
          ContextEntry::Marker(m) if *m == ex_a_s => true,
          _ => false,
        })
        .unwrap();
      info!("Unsolved: {:?}", after_marker.unsolved_existentials());

      let mut tau =
        app_ctx(&after_marker, &Type::Arrow(Box::new(ex_a), Box::new(ex_b)))?;
      for name in after_marker.unsolved_existentials() {
        warn!("substituting {tau:?} looking for {name}");
        tau = tau.subst(&name, &Type::Var(ex_gen.clone()));
      }

      Ok((Type::Polytype(ex_gen, Box::new(tau)), before_marker))
    }

    // // →I⇒ rule
    // // > Rule →I⇒ corresponds to Decl→I⇒, one of the guessing rules, so we
    // // > create new existential variables ^α (for the function domain) and ^β (for
    // // > the codomain) and check the function body against ^β.  As in ∀App, we do
    // // > not place a marker before ^α, because ^α and ^β appear in the output type
    // // > (λx. e ⇒ ^α → ^β).
    // Term::Lam(x, e) => {
    //   let ex_a_s = gensym_existential();
    //   let ex_b_s = gensym_existential();
    //   let ex_a = Type::Existential(ex_a_s.clone());
    //   let ex_b = Type::Existential(ex_b_s.clone());
    //   let aug_ctx = ctx.add(vec![
    //     ContextEntry::ExistentialVar(ex_a_s.clone()),
    //     ContextEntry::ExistentialVar(ex_b_s.clone()),
    //     ContextEntry::TermAnnot(x.clone(), ex_a.clone()),
    //   ]);

    //   let wtf_ctx = typecheck(&aug_ctx, &e, &ex_b)?;
    //   Ok((Type::Arrow(Box::new(ex_a), Box::new(ex_b)), ctx.clone()))
    // }

    // →E rule
    Term::App(e1, e2) => {
      let (ty_a, ctx_theta) = synthesize(ctx, e1)?;
      let app_a = app_ctx(&ctx_theta, &ty_a)?;
      let (ty_c, ctx_delta) = app_synthesize(&ctx_theta, &app_a, &e2)?;
      Ok((ty_c, ctx_delta))
    }
  }
}

#[cfg_attr(feature = "trace_execution", trace)]
pub fn app_synthesize(
  ctx: &Context,
  fun_ty: &Type,
  term: &Term,
) -> Result<(Type, Context)> {
  match (fun_ty, term) {
    // →App rule
    (Type::Arrow(ty_a, ty_c), e) => {
      let out_ctx = typecheck(ctx, e, ty_a)?;
      Ok((*ty_c.clone(), out_ctx))
    }

    // ∀App rule
    (Type::Polytype(a, ty_a), e) => {
      let ex_s = gensym_existential();
      let ex = ContextEntry::ExistentialVar(ex_s.clone());
      let aug_ctx = ctx.add(vec![ex.clone()]);

      let aug_ty = ty_a.subst(&a, &Type::Existential(ex_s));
      let (ty, ctx_delta) = app_synthesize(&aug_ctx, &aug_ty, e)?;

      Ok((ty, ctx_delta))
    }

    // âApp rule
    (Type::Existential(a), e) if ctx.has_existential(a) => {
      let (ex_a1_s, ex_a2_s, ctx_gamma_aug) =
        ctx.split_existential_function(a)?;

      let ex_a1 = Type::Existential(ex_a1_s.clone());
      let ctx_delta = typecheck(&ctx_gamma_aug, e, &ex_a1)?;

      let ex_a2 = Type::Existential(ex_a2_s.clone());
      Ok((ex_a2, ctx_delta))
    }

    _ => bail!("trying to appSynthesize with a non-function type"),
  }
}
