use anyhow::{bail, ensure, Result};

use crate::data_debruijn::{
  Context, ContextEntry, ContextIndex, Monotype, Term, Type, Visitor,
};
use crate::DEPTH;

// Figure 8. Applying a context, as a substitution, to a type

#[trace]
pub fn app_ctx(ctx: &Context, ty: &Type) -> Result<Type> {
  match ty {
    Type::Unit => Ok(Type::Unit),
    Type::Var(s) => Ok(Type::Var(s.clone())),

    Type::Existential(a) => match ctx.get_existential(a) {
      Some(Some(m)) => Ok(m.into_poly()),
      Some(None) => Ok(Type::Existential(a.clone())),
      None => bail!("existential variable {a:?} doesn't exist in context"),
    },

    // TODO: Is this right?
    // Type::Polytype(t) => {
    //   let t = app_ctx(ctx, t)?;
    //   Ok(Type::Polytype(Box::new(t)))
    // }
    Type::Polytype(t) => {
      let (aug_ctx, _) = ctx.add(ContextEntry::ExistentialVar)?;
      let t = app_ctx(&aug_ctx, t)?;
      Ok(Type::Polytype(Box::new(t)))
    }

    Type::Arrow(a, b) => Ok(Type::Arrow(
      Box::new(app_ctx(ctx, a)?),
      Box::new(app_ctx(ctx, b)?),
    )),
  }
}

// Figure 9. Algorithmic subtyping

/// Under input context Γ , type A is a subtype of B, with output context ∆
pub fn subtype(ctx: &Context, left: &Type, right: &Type) -> Result<Context> {
  match (left, right) {
    // <:Unit rule
    (Type::Unit, Type::Unit) => Ok(ctx.clone()),

    // <:Var rule
    (Type::Var(x), Type::Var(y)) if x == y && ctx.get_type(x).is_some() => {
      Ok(ctx.clone())
    }

    // <:Exvar rule
    (Type::Existential(x), Type::Existential(y))
      if x == y && ctx.get_existential(x).is_some() =>
    {
      Ok(ctx.clone())
    }

    // <:InstantiateL
    (Type::Existential(a) , ty_a) if ctx.get_existential(a).is_some() => {
      instantiate_left(ctx, a, ty_a)
    }

    // <:InstantiateR
    (ty_a, Type::Existential(a)) if ctx.get_existential(a).is_some() => {
      instantiate_right (ctx, ty_a, a)
    }

    (Type::Unit, Type::Var(_)) => todo!("wtf? {left:?} {right:?}"),
    (Type::Unit, Type::Polytype(_)) => todo!(),
    (Type::Unit, Type::Arrow(_, _)) => todo!(),
    (Type::Var(_), Type::Unit) => todo!(),
    (Type::Var(_), Type::Var(_)) => todo!(),
    (Type::Var(_), Type::Existential(_)) => todo!(),
    (Type::Var(_), Type::Polytype(_)) => todo!(),
    (Type::Var(_), Type::Arrow(_, _)) => todo!(),
    (Type::Existential(_), Type::Unit) => todo!(),
    (Type::Existential(_), Type::Var(_)) => todo!(),
    (Type::Existential(_), Type::Existential(_)) => todo!(),
    (Type::Existential(_), Type::Polytype(_)) => todo!(),
    (Type::Existential(_), Type::Arrow(_, _)) => todo!(),
    (Type::Polytype(_), Type::Unit) => todo!(),
    (Type::Polytype(_), Type::Var(_)) => todo!(),
    (Type::Polytype(_), Type::Existential(_)) => todo!(),
    (Type::Polytype(_), Type::Polytype(_)) => todo!(),
    (Type::Polytype(_), Type::Arrow(_, _)) => todo!(),
    (Type::Arrow(_, _), Type::Unit) => todo!(),
    (Type::Arrow(_, _), Type::Var(_)) => todo!(),
    (Type::Arrow(_, _), Type::Existential(_)) => todo!(),
    (Type::Arrow(_, _), Type::Polytype(_)) => todo!(),
    (Type::Arrow(_, _), Type::Arrow(_, _)) => todo!(),

    _ => bail!("subtyping relation failed between {left:?} and {right:?} (ctx = {ctx:?})"),
  }
}

// Figure 10. Instantiation

pub fn instantiate_left(
  ctx: &Context,
  a: &ContextIndex,
  ty_a: &Type,
) -> Result<Context> {
  match ty_a {
    // InstLReach rule
    Type::Existential(b)
      if ctx.get_existential(a).is_some()
        && ctx.get_existential(b).is_some() =>
    {
      let mut new_ctx = ctx.clone();
      {
        let b_entry = new_ctx
          .vector
          .get_mut(b.debruijn_level)
          .expect("should exist");

        ensure!(
          matches!(b_entry, ContextEntry::ExistentialVar),
          "not an existential variable"
        );

        *b_entry =
          ContextEntry::ExistentialSolved(Monotype::Existential(a.to_owned()));
      }

      Ok(new_ctx)
    }

    Type::Existential(_b) => todo!(),
    Type::Unit => todo!(),
    Type::Var(_) => todo!(),
    Type::Polytype(_) => todo!(),
    Type::Arrow(_, _) => todo!(),
  }
}

pub fn instantiate_right(
  ctx: &Context,
  ty_a: &Type,
  a: &ContextIndex,
) -> Result<Context> {
  match ty_a {
    // InstRSolve rule
    ty if ty.is_mono() => {
      let mut new_ctx = ctx.clone();
      {
        let a_entry = new_ctx
          .vector
          .get_mut(a.debruijn_level)
          .expect("should exist");

        ensure!(
          matches!(a_entry, ContextEntry::ExistentialVar),
          "not an existential variable"
        );

        *a_entry = ContextEntry::ExistentialSolved(ty.try_into_mono().unwrap());
      }
      Ok(new_ctx)
    }

    Type::Unit => todo!("a={a:?}, ty_a={ty_a:?}, ctx={ctx:?}"),
    Type::Var(_) => todo!(),
    Type::Existential(_) => todo!(),
    Type::Polytype(_) => todo!(),
    Type::Arrow(_, _) => todo!(),
  }
}

// Figure 11. Algorithmic typing

#[trace]
pub fn typecheck(ctx: &Context, term: &Term, ty: &Type) -> Result<Context> {
  match (term, ty) {
    // 1I rule
    (Term::Unit, Type::Unit) => Ok(ctx.clone()),

    // ∀I rule
    (_e, Type::Polytype(_t)) => todo!(),

    // →I rule
    (Term::Lam(_e), Type::Arrow(_ty_a, _ty_b)) => {
      todo!()
    }

    // Sub rule
    (term, ty) => {
      let (ty_a, ctx_theta) = synthesize(ctx, term)?;
      let a = app_ctx(&ctx_theta, &ty_a)?;
      let b = app_ctx(&ctx_theta, ty)?;
      let ctx_delta = subtype(&ctx_theta, &a, &b)?;
      Ok(ctx_delta)
    }
  }
}

#[trace]
pub fn synthesize(ctx: &Context, term: &Term) -> Result<(Type, Context)> {
  match term {
    // 1I⇒ rule
    Term::Unit => Ok((Type::Unit, ctx.clone())),

    // Var rule
    Term::Var(index) => {
      let ty = match ctx.get_type(index) {
        Some(v) => v,
        None => bail!("no TermAnnot at index {index:?} of context: {ctx:?}"),
      };
      Ok((ty, ctx.clone()))
    }

    // →E rule
    Term::App(e1, e2) => {
      let (ty_a, ctx_theta) = synthesize(ctx, e1)?;
      let app_a = app_ctx(&ctx_theta, &ty_a)?;
      let (ty_c, ctx_delta) = app_synthesize(&ctx_theta, &app_a, &e2)?;
      Ok((ty_c, ctx_delta))
    }

    // →I⇒' rule
    Term::Lam(e) => {
      let (aug_ctx, marker_idx) = ctx.add(ContextEntry::Marker)?;
      let (aug_ctx_at_a, ex_a_idx) =
        aug_ctx.add(ContextEntry::ExistentialVar)?;
      let (aug_ctx_at_b, ex_b_idx) =
        aug_ctx_at_a.add(ContextEntry::ExistentialVar)?;

      let ex_a = Type::Existential(ex_a_idx.clone());
      let ex_b = Type::Existential(ex_b_idx.clone());

      let (aug_ctx, _annot_idx) =
        aug_ctx_at_b.add(ContextEntry::TermAnnot(ex_a.clone()))?;

      struct V(usize, usize);
      impl Visitor for V {
        fn visit_index(&self, index: &mut ContextIndex) -> Result<()> {
          index.debruijn_index += self.0;
          index.debruijn_level = self.1 - 1 - index.debruijn_index;
          Ok(())
        }
      }

      let ex_a_2 = ex_b.transform_with_visitor(&V(2, aug_ctx.size()))?;
      let ex_b_2 = ex_b.transform_with_visitor(&V(1, aug_ctx.size()))?;
      let e_2 = e.transform_with_visitor(&V(0, aug_ctx.size()))?;

      let wtf_ctx = typecheck(&aug_ctx, &e_2, &ex_b_2)?;

      let (before_marker, after_marker) = wtf_ctx.split_at(&marker_idx);

      // Bump the levels down since we're splitting teh array
      struct V2(usize, usize);
      impl Visitor for V2 {
        fn visit_index(&self, index: &mut ContextIndex) -> Result<()> {
          index.debruijn_level -= self.0;
          index.debruijn_index = self.1 - 1 - index.debruijn_level;
          Ok(())
        }
      }
      let after_marker =
        after_marker.transform_with_visitor(&V2(1, after_marker.size()))?;

      // τ=[∆'](â→β)
      let tau = Type::Arrow(Box::new(ex_a_2), Box::new(ex_b_2));
      let tau = tau.transform_with_visitor(&V2(1, after_marker.size()))?;
      let tau = app_ctx(&after_marker, &tau)?;

      // let mut before_marker = before_marker;
      // before_marker.vector.push_back(ContextEntry::ExistentialVar);

      Ok((Type::Polytype(Box::new(tau)), before_marker))
    }

    Term::Annot(_, _) => todo!(),
  }
}

pub fn app_synthesize(
  ctx: &Context,
  fun_ty: &Type,
  term: &Term,
) -> Result<(Type, Context)> {
  match (fun_ty, term) {
    // →App rule
    (Type::Arrow(ty_a, ty_c), e) => {
      let out_ctx = typecheck(ctx, e, ty_a)?;
      Ok((*ty_c.clone(), out_ctx))
    }

    // ∀App rule
    (Type::Polytype(ty_a), e) => {
      let (aug_ctx, a_idx) = ctx.add(ContextEntry::ExistentialVar)?;
      let aug_ty = ty_a.subst(&a_idx, Type::Existential(a_idx.clone()));
      let (ty, ctx_delta) = app_synthesize(&aug_ctx, &aug_ty, e)?;

      Ok((ty, ctx_delta))
    }

    (Type::Unit, Term::Unit) => todo!(),
    (Type::Unit, Term::Var(_)) => todo!(),
    (Type::Unit, Term::Lam(_)) => todo!(),
    (Type::Unit, Term::App(_, _)) => todo!(),
    (Type::Unit, Term::Annot(_, _)) => todo!(),
    (Type::Var(_), Term::Unit) => todo!(),
    (Type::Var(_), Term::Var(_)) => todo!(),
    (Type::Var(_), Term::Lam(_)) => todo!(),
    (Type::Var(_), Term::App(_, _)) => todo!(),
    (Type::Var(_), Term::Annot(_, _)) => todo!(),
    (Type::Existential(_), Term::Unit) => todo!(),
    (Type::Existential(_), Term::Var(_)) => todo!(),
    (Type::Existential(_), Term::Lam(_)) => todo!(),
    (Type::Existential(_), Term::App(_, _)) => todo!(),
    (Type::Existential(_), Term::Annot(_, _)) => todo!(),

    _ => bail!("trying to appSynthesize with a non-function type"),
  }
}
