use std::{fmt, hash::Hash};

use anyhow::{bail, Result};
use im::{HashSet, Vector};

use crate::gensym::gensym_existential;

#[derive(Clone)]
pub enum Term {
  Unit,
  Var(String),
  Lam(String, Box<Term>),
  App(Box<Term>, Box<Term>),
  Annot(Box<Term>, Type),
}

impl fmt::Debug for Term {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::Unit => write!(f, "unit"),
      Self::Var(arg0) => write!(f, "{}", arg0),
      Self::Lam(arg0, arg1) => write!(f, "(λ{}.{:?})", arg0, arg1),
      Self::App(arg0, arg1) => write!(f, "({:?} · {:?})", arg0, arg1),
      Self::Annot(arg0, arg1) => write!(f, "({:?} : {:?})", arg0, arg1),
    }
  }
}

#[derive(Clone)]
pub enum Type {
  Unit,
  Var(String),
  Existential(String),
  Polytype(String, Box<Type>),
  Arrow(Box<Type>, Box<Type>),
}

impl fmt::Debug for Type {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::Unit => write!(f, "𝟙"),
      Self::Var(arg0) => write!(f, "{}", arg0),
      Self::Existential(arg0) => write!(f, "{}", arg0),
      Self::Polytype(arg0, arg1) => write!(f, "(∀ {} . {:?})", arg0, arg1),
      Self::Arrow(arg0, arg1) => write!(f, "({:?} -> {:?})", arg0, arg1),
    }
  }
}

impl Type {
  pub fn into_mono(&self) -> Option<Monotype> {
    match self {
      Type::Unit => Some(Monotype::Unit),
      Type::Var(x) => Some(Monotype::Var(x.clone())),
      Type::Existential(x) => Some(Monotype::Existential(x.clone())),
      Type::Polytype(_, _) => None,
      Type::Arrow(a, b) => Some(Monotype::Arrow(
        Box::new(a.into_mono()?),
        Box::new(b.into_mono()?),
      )),
    }
  }

  #[inline]
  pub fn is_mono(&self) -> bool {
    self.into_mono().is_some()
  }
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum FreeVar {
  Var(String),
  Existential(String),
}

impl Type {
  pub fn free_vars(&self) -> HashSet<FreeVar> {
    fn free_vars_with_context(ctx: &Context, ty: &Type) -> HashSet<FreeVar> {
      match ty {
        Type::Unit => HashSet::default(),
        Type::Var(x) if ctx.lookup_type(x).is_some() => HashSet::default(),
        Type::Var(x) => [FreeVar::Var(x.to_owned())].into_iter().collect(),
        Type::Existential(x) if ctx.lookup_existential(x).is_some() => {
          HashSet::default()
        }
        Type::Existential(x) => {
          [FreeVar::Existential(x.to_owned())].into_iter().collect()
        }
        Type::Polytype(x, ty_a) => {
          let new_ctx =
            ctx.add(vec![ContextEntry::ExistentialVar(x.to_owned())]);
          free_vars_with_context(&new_ctx, &ty_a)
        }
        Type::Arrow(ty_a, ty_b) => {
          let a_vars = free_vars_with_context(&ctx, &ty_a);
          let b_vars = free_vars_with_context(&ctx, &ty_b);
          a_vars.union(b_vars)
        }
      }
    }

    free_vars_with_context(&Context::default(), self)
  }

  pub fn subst(&self, var: &str, replacement: &Type) -> Type {
    match self {
      Type::Unit => Type::Unit,
      Type::Var(n) if n == var => replacement.clone(),
      Type::Var(n) => Type::Var(n.clone()),
      Type::Existential(s) if s == var => replacement.clone(),
      Type::Existential(s) => Type::Existential(s.clone()),
      Type::Polytype(a, t) => {
        Type::Polytype(a.clone(), Box::new(t.subst(var, replacement)))
      }
      Type::Arrow(a, b) => Type::Arrow(
        Box::new(a.subst(var, replacement)),
        Box::new(b.subst(var, replacement)),
      ),
    }
  }
}

#[derive(Clone)]
pub enum Monotype {
  Unit,
  Var(String),
  Existential(String),
  Arrow(Box<Monotype>, Box<Monotype>),
}

impl fmt::Debug for Monotype {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::Unit => write!(f, "𝟙"),
      Self::Var(arg0) => write!(f, "{}", arg0),
      Self::Existential(arg0) => write!(f, "{}", arg0),
      Self::Arrow(arg0, arg1) => write!(f, "({arg0:?} -> {arg1:?})"),
    }
  }
}

impl Monotype {
  pub fn into_poly(&self) -> Type {
    match self {
      Monotype::Unit => Type::Unit,
      Monotype::Var(x) => Type::Var(x.clone()),
      Monotype::Existential(x) => Type::Existential(x.clone()),
      Monotype::Arrow(a, b) => {
        Type::Arrow(Box::new(a.into_poly()), Box::new(b.into_poly()))
      }
    }
  }
}

#[derive(Clone)]
pub enum ContextEntry {
  TypeVar(String),
  TermAnnot(String, Type),
  ExistentialVar(String),
  ExistentialSolved(String, Monotype),
  Marker(String),
}

impl fmt::Debug for ContextEntry {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::TypeVar(arg0) => write!(f, "{}", arg0),
      Self::TermAnnot(arg0, arg1) => write!(f, "{} : {:?}", arg0, arg1),
      Self::ExistentialVar(arg0) => write!(f, "{}", arg0),
      Self::ExistentialSolved(arg0, arg1) => write!(f, "{} ≔ {:?}", arg0, arg1),
      Self::Marker(arg0) => write!(f, "▶{}", arg0),
    }
  }
}

#[derive(Debug, Clone)]
pub enum CompleteContextEntry {
  TypeVar(String),
  TermAnnot(String, Type),
  ExistentialSolved(String, Monotype),
  Marker(String),
}

#[derive(Clone, Default)]
pub struct Context(pub(crate) Vector<ContextEntry>);

impl fmt::Debug for Context {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    f.write_str("Γ")?;
    for rule in self.0.iter() {
      f.write_str(", ")?;
      rule.fmt(f)?;
    }
    Ok(())
  }
}

impl Context {
  pub fn add(&self, entries: Vec<ContextEntry>) -> Context {
    let mut new_ctx = self.clone();
    for entry in entries {
      new_ctx.0.push_back(entry);
    }
    new_ctx
  }

  /// Looks up a polytype by name, also returning an index
  pub fn lookup_type(&self, name: impl AsRef<str>) -> Option<(usize, Type)> {
    self
      .0
      .iter()
      .enumerate()
      .find_map(|(i, entry)| match entry {
        ContextEntry::TermAnnot(n, t) if n == name.as_ref() => {
          Some((i, t.clone()))
        }
        _ => None,
      })
  }

  #[inline]
  pub fn has_type(&self, name: impl AsRef<str>) -> bool {
    self.lookup_type(name).is_some()
  }

  /// Looks up an existential variable by name
  /// - Returns Some(Some(t)) if solved
  /// - Returns Some(None) if exists but unsolved
  /// - Returns None if not found
  pub fn lookup_existential(
    &self,
    name: impl AsRef<str>,
  ) -> Option<(usize, Option<Monotype>)> {
    self
      .0
      .iter()
      .enumerate()
      .find_map(|(i, entry)| match entry {
        ContextEntry::ExistentialVar(n) if n == name.as_ref() => {
          Some((i, None))
        }
        ContextEntry::ExistentialSolved(n, t) if n == name.as_ref() => {
          Some((i, Some(t.clone())))
        }
        _ => None,
      })
  }

  #[inline]
  pub fn has_existential(&self, name: impl AsRef<str>) -> bool {
    self.lookup_existential(name).is_some()
  }

  /// Returns a list of names of unsolved existentials
  pub fn unsolved_existentials(&self) -> HashSet<String> {
    let mut unsolved = HashSet::new();

    for entry in self.0.iter() {
      match entry {
        ContextEntry::ExistentialVar(n) => {
          unsolved.insert(n.clone());
        }
        _ => {}
      }
    }

    unsolved
  }

  /// Returns (before, after)
  pub fn split_by<P>(&self, p: P) -> Option<(Context, Context)>
  where
    P: Fn(&ContextEntry) -> bool,
  {
    let (before, after) = {
      let idx = self.0.iter().position(p)?;
      self.0.clone().split_at(idx)
    };

    Some((Context(before), Context(after)))
  }

  pub fn unsplit(
    left: &Context,
    center: ContextEntry,
    right: &Context,
  ) -> Context {
    let mut res = left.clone();
    res.0.push_back(center);
    res.0.extend(right.0.clone().into_iter());
    res
  }

  pub fn split_existential_function(
    &self,
    var: &str,
  ) -> Result<(String, String, Context)> {
    let mut ctx = self.clone();
    let idx = match ctx.lookup_existential(var) {
      Some((idx, _)) => idx,
      None => bail!("wtf?"),
    };

    let ex_a1_s = gensym_existential();
    let ex_a2_s = gensym_existential();
    ctx.0[idx] = ContextEntry::ExistentialSolved(
      var.to_owned(),
      Monotype::Arrow(
        Box::new(Monotype::Existential(ex_a1_s.to_owned())),
        Box::new(Monotype::Existential(ex_a2_s.to_owned())),
      ),
    );
    ctx
      .0
      .insert(idx, ContextEntry::ExistentialVar(ex_a1_s.to_owned()));
    ctx
      .0
      .insert(idx, ContextEntry::ExistentialVar(ex_a2_s.to_owned()));

    Ok((ex_a1_s, ex_a2_s, ctx))
  }
}
