use std::fmt::{self};

use anyhow::Result;
use im::Vector;

use crate::{data, gensym::gensym_type};

/// A lambda calculus term.
#[derive(Clone)]
pub enum Term {
  /// Unit type
  Unit,

  /// Variable, with a reference into the context.
  /// The entry pointed to by this index MUST be a TypeVar.
  Var(ContextIndex),

  /// Lambda abstraction.
  Lam(Box<Term>),

  /// Lambda application.
  App(Box<Term>, Box<Term>),

  /// Type annotation
  Annot(Box<Term>, Type),
}

impl fmt::Debug for Term {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::Unit => write!(f, "unit"),
      Self::Var(arg0) => write!(f, "{:?}", arg0),
      Self::Lam(arg1) => write!(f, "(λ.{:?})", arg1),
      Self::App(arg0, arg1) => write!(f, "({:?} · {:?})", arg0, arg1),
      Self::Annot(arg0, arg1) => write!(f, "({:?} : {:?})", arg0, arg1),
    }
  }
}

impl Term {
  pub fn transform_with_visitor<V: Visitor>(
    &self,
    visitor: &V,
  ) -> Result<Term> {
    let mut expr = self.clone();
    visitor.visit_term_before(&mut expr)?;
    let mut expr = match expr {
      Term::Var(index) => {
        let mut index = index.clone();
        visitor.visit_index(&mut index)?;
        Term::Var(index)
      }
      Term::Lam(_) => todo!(),
      Term::App(_, _) => todo!(),
      Term::Annot(_, _) => todo!(),
      _ => expr,
    };
    visitor.visit_term_after(&mut expr)?;
    Ok(expr)
  }
}

#[derive(Clone)]
pub enum Type {
  Unit,
  Var(ContextIndex),
  Existential(ContextIndex),
  Polytype(Box<Type>),
  Arrow(Box<Type>, Box<Type>),
}

impl fmt::Debug for Type {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::Unit => write!(f, "𝟙"),
      Self::Var(arg0) => write!(f, "{:?}", arg0),
      Self::Existential(arg0) => write!(f, "ex{:?}", arg0),
      Self::Polytype(arg1) => write!(f, "(∀.{:?})", arg1),
      Self::Arrow(arg0, arg1) => write!(f, "({:?} -> {:?})", arg0, arg1),
    }
  }
}

impl Type {
  /// Attempt to turn this polytype into a monotype.
  ///
  /// This fails if any sub-expression of this type contains a polytype expression.
  pub fn try_into_mono(&self) -> Option<Monotype> {
    match self {
      Type::Unit => Some(Monotype::Unit),
      Type::Var(x) => Some(Monotype::Var(x.clone())),
      Type::Existential(x) => Some(Monotype::Existential(x.clone())),

      // Polytypes cannot be converted to monotypes
      Type::Polytype(_) => None,

      Type::Arrow(a, b) => Some(Monotype::Arrow(
        Box::new(a.try_into_mono()?),
        Box::new(b.try_into_mono()?),
      )),
    }
  }

  #[inline]
  pub fn is_mono(&self) -> bool {
    self.try_into_mono().is_some()
  }

  pub fn subst(&self, before: &ContextIndex, after: Type) -> Type {
    match self {
      Type::Unit => Type::Unit,
      Type::Var(index) if index.debruijn_level == before.debruijn_level => {
        after
      }
      Type::Var(index) => Type::Var(index.clone()),
      Type::Existential(index)
        if index.debruijn_level == before.debruijn_level =>
      {
        after
      }
      Type::Existential(index) => Type::Existential(index.clone()),
      Type::Polytype(_) => todo!(),
      Type::Arrow(a, b) => Type::Arrow(
        Box::new(a.subst(before, after.clone())),
        Box::new(b.subst(before, after.clone())),
      ),
    }
  }

  pub fn transform_with_visitor<V: Visitor>(
    &self,
    visitor: &V,
  ) -> Result<Type> {
    let mut ty = self.clone();
    visitor.visit_type_before(&mut ty)?;
    let mut ty = match ty {
      Type::Var(_) => todo!(),
      Type::Existential(index) => {
        let mut index = index.clone();
        visitor.visit_index(&mut index)?;
        Type::Existential(index)
      }
      Type::Polytype(_) => todo!(),
      Type::Arrow(a, b) => {
        let a = a.transform_with_visitor(visitor)?;
        let b = b.transform_with_visitor(visitor)?;
        Type::Arrow(Box::new(a), Box::new(b))
      }
      _ => ty,
    };
    visitor.visit_type_after(&mut ty)?;
    Ok(ty)
  }
}

#[derive(Debug, Clone)]
pub enum Monotype {
  Unit,
  Var(ContextIndex),
  Existential(ContextIndex),
  Arrow(Box<Monotype>, Box<Monotype>),
}

impl Monotype {
  pub fn into_poly(&self) -> Type {
    match self {
      Monotype::Unit => Type::Unit,
      Monotype::Var(x) => Type::Var(x.clone()),
      Monotype::Existential(x) => Type::Existential(x.clone()),
      Monotype::Arrow(a, b) => {
        Type::Arrow(Box::new(a.into_poly()), Box::new(b.into_poly()))
      }
    }
  }

  pub fn transform_with_visitor<V: Visitor>(
    &self,
    visitor: &V,
  ) -> Result<Monotype> {
    let mut ty = self.clone();
    visitor.visit_monotype_before(&mut ty)?;
    let mut ty = match ty {
      Monotype::Var(_) => todo!(),
      Monotype::Existential(index) => {
        let mut index = index.clone();
        visitor.visit_index(&mut index)?;
        Monotype::Existential(index)
      }
      Monotype::Arrow(_, _) => todo!(),
      _ => ty,
    };
    visitor.visit_monotype_after(&mut ty)?;
    Ok(ty)
  }
}

#[derive(Clone)]
pub enum ContextEntry {
  TypeVar,
  TermAnnot(Type),
  ExistentialVar,
  ExistentialSolved(Monotype),
  Marker,
}

impl fmt::Debug for ContextEntry {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    match self {
      Self::TypeVar => write!(f, "Ty"),
      Self::TermAnnot(arg0) => f.debug_tuple("Annot").field(arg0).finish(),
      Self::ExistentialVar => write!(f, "ExVar"),
      Self::ExistentialSolved(arg0) => {
        f.debug_tuple("ExSolv").field(arg0).finish()
      }
      Self::Marker => write!(f, "▶"),
    }
  }
}

#[derive(Debug, Clone)]
pub enum CompleteContextEntry {
  TypeVar(String),
  TermAnnot(String, Type),
  ExistentialSolved(String, Monotype),
  Marker(String),
}

#[derive(Clone)]
pub struct Context {
  // TODO: Make this not pub(crate)
  pub(crate) vector: Vector<ContextEntry>,
}

impl fmt::Debug for Context {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    write!(f, "Γ{:?}", self.vector)
  }
}

impl Default for Context {
  fn default() -> Self {
    let vector = Vector::default();

    Context { vector }
  }
}

impl Context {
  #[inline]
  pub fn size(&self) -> usize {
    self.vector.len()
  }

  pub fn add(&self, item: ContextEntry) -> Result<(Context, ContextIndex)> {
    let idx = self.vector.len();
    struct V;
    impl Visitor for V {
      fn visit_index(&self, index: &mut ContextIndex) -> Result<()> {
        index.debruijn_index += 1;
        Ok(())
      }
    }
    let mut new_ctx = self.transform_with_visitor(&V)?;
    new_ctx.vector.push_back(item);

    let idx = ContextIndex {
      debruijn_index: 0,
      debruijn_level: idx,
    };
    Ok((new_ctx, idx))
  }

  pub fn get<'a>(&'a self, index: &ContextIndex) -> Option<&ContextEntry> {
    let item = self.vector.get(index.debruijn_level)?;
    Some(item)
  }

  pub fn get_existential(
    &self,
    index: &ContextIndex,
  ) -> Option<Option<Monotype>> {
    let item = self.get(index)?;

    match item {
      ContextEntry::ExistentialSolved(t) => Some(Some(t.clone())),
      ContextEntry::ExistentialVar => Some(None),
      _ => None,
    }
  }

  pub fn get_type(&self, index: &ContextIndex) -> Option<Type> {
    let item = self.get(index)?;

    match item {
      ContextEntry::TermAnnot(t) => Some(t.clone()),
      _ => None,
    }
  }

  pub fn unsolved_existentials(&self) -> Vec<ContextIndex> {
    let mut unsolved = Vec::new();
    for (idx, item) in self.vector.iter().enumerate() {
      match item {
        ContextEntry::ExistentialVar => {
          let index = ContextIndex {
            debruijn_level: idx,
            debruijn_index: self.vector.len() - idx,
          };
          unsolved.push(index)
        }
        _ => {}
      }
    }
    unsolved
  }

  /// Splits the context
  pub fn split_at(&self, index: &ContextIndex) -> (Context, Context) {
    let (before, mut after) =
      self.vector.clone().split_at(index.debruijn_level);
    after.pop_front();

    (Context { vector: before }, Context { vector: after })
  }

  pub fn is_valid(&self) -> bool {
    for item in self.vector.iter() {
      match item {
        ContextEntry::TermAnnot(_) => todo!(),
        ContextEntry::ExistentialSolved(_) => todo!(),
        _ => {}
      }
    }
    true
  }

  pub fn ensure_validity(&self) -> bool {
    fn ensure_item_validity(ctx: &Context, item: &ContextEntry) -> bool {
      match item {
        ContextEntry::TermAnnot(ty) => ensure_type_validity(ctx, ty),
        ContextEntry::ExistentialSolved(ty) => {
          ensure_monotype_validity(ctx, ty)
        }
        _ => true,
      }
    }

    fn ensure_type_validity(ctx: &Context, ty: &Type) -> bool {
      match ty {
        Type::Var(idx) => ensure_index_validity(ctx, idx),
        Type::Existential(idx) => ensure_index_validity(ctx, idx),
        Type::Polytype(ty) => ensure_type_validity(ctx, ty),
        Type::Arrow(ty1, ty2) => {
          ensure_type_validity(ctx, ty1) && ensure_type_validity(ctx, ty2)
        }
        _ => true,
      }
    }

    fn ensure_monotype_validity(ctx: &Context, ty: &Monotype) -> bool {
      match ty {
        Monotype::Var(idx) => ensure_index_validity(ctx, idx),
        Monotype::Existential(idx) => ensure_index_validity(ctx, idx),
        Monotype::Arrow(ty1, ty2) => {
          ensure_monotype_validity(ctx, ty1)
            && ensure_monotype_validity(ctx, ty2)
        }
        _ => true,
      }
    }

    fn ensure_index_validity(ctx: &Context, idx: &ContextIndex) -> bool {
      idx.debruijn_index + idx.debruijn_level == ctx.vector.len()
    }

    self
      .vector
      .iter()
      .all(|item| ensure_item_validity(self, item))
  }

  pub fn transform_with_visitor<V: Visitor>(
    &self,
    visitor: &V,
  ) -> Result<Context> {
    Ok(Context {
      vector: self
        .vector
        .iter()
        .map(|entry| {
          let mut entry = entry.clone();
          visitor.visit_context_entry_before(&mut entry)?;
          let mut entry = match entry {
            ContextEntry::TermAnnot(ty) => {
              ContextEntry::TermAnnot(ty.transform_with_visitor(visitor)?)
            }
            ContextEntry::ExistentialSolved(ty) => {
              ContextEntry::ExistentialSolved(
                ty.transform_with_visitor(visitor)?,
              )
            }
            _ => entry,
          };
          visitor.visit_context_entry_after(&mut entry)?;
          Ok(entry)
        })
        .collect::<Result<_>>()?,
    })
  }
}

/// An index into the context.
#[derive(Clone, PartialEq, Eq)]
pub struct ContextIndex {
  pub(crate) debruijn_index: usize,
  pub(crate) debruijn_level: usize,
}

impl fmt::Debug for ContextIndex {
  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
    write!(f, "#{}/{}", self.debruijn_level, self.debruijn_index)
  }
}

/* Convert                                                                    */

pub fn convert_term(term: &data::Term) -> Term {
  fn convert_term_with_context(
    ctx: &data::Context,
    ctx2: &Context,
    term: &data::Term,
  ) -> Term {
    match term {
      data::Term::Unit => Term::Unit,

      data::Term::Var(name) => {
        println!("Looking up name {name}");
        let (idx, _) = ctx.lookup_type(name).unwrap();

        let index = ContextIndex {
          debruijn_index: 0,
          debruijn_level: idx,
        };
        Term::Var(index)
      }

      data::Term::Lam(arg, body) => {
        let ty = gensym_type();
        let aug_ctx = ctx.add(vec![data::ContextEntry::TermAnnot(
          arg.to_owned(),
          data::Type::Var(ty.clone()),
        )]);
        let body = convert_term_with_context(&aug_ctx, ctx2, body);
        Term::Lam(Box::new(body))
      }

      data::Term::App(func, arg) => {
        let func = convert_term_with_context(ctx, ctx2, &func);
        let arg = convert_term_with_context(ctx, ctx2, &arg);
        Term::App(Box::new(func), Box::new(arg))
      }

      data::Term::Annot(term, ty) => {
        let term = convert_term_with_context(ctx, ctx2, &term);
        let ty = convert_ty_with_context(ctx, ctx2, &ty);
        Term::Annot(Box::new(term), ty)
      }
    }
  }

  fn convert_ty_with_context(
    _ctx: &data::Context,
    _ctx2: &Context,
    ty: &data::Type,
  ) -> Type {
    match ty {
      data::Type::Unit => Type::Unit,

      data::Type::Var(_) => todo!(),
      data::Type::Existential(_) => todo!(),
      data::Type::Polytype(_, _) => todo!(),
      data::Type::Arrow(_, _) => todo!(),
    }
  }

  convert_term_with_context(
    &data::Context::default(),
    &Context::default(),
    term,
  )
}

pub trait Visitor {
  fn visit_context_entry_before(&self, expr: &mut ContextEntry) -> Result<()> {
    Ok(())
  }
  fn visit_context_entry_after(&self, expr: &mut ContextEntry) -> Result<()> {
    Ok(())
  }
  fn visit_type_before(&self, expr: &mut Type) -> Result<()> {
    Ok(())
  }
  fn visit_type_after(&self, expr: &mut Type) -> Result<()> {
    Ok(())
  }
  fn visit_monotype_before(&self, expr: &mut Monotype) -> Result<()> {
    Ok(())
  }
  fn visit_monotype_after(&self, expr: &mut Monotype) -> Result<()> {
    Ok(())
  }
  fn visit_term_before(&self, expr: &mut Term) -> Result<()> {
    Ok(())
  }
  fn visit_term_after(&self, expr: &mut Term) -> Result<()> {
    Ok(())
  }
  fn visit_index(&self, index: &mut ContextIndex) -> Result<()> {
    Ok(())
  }
}
